from .quinfig import Quinfig
from .common.sweep import QuinSweep
from .common.argparse import QuinineArgumentParser
from .common.cerberus import tstring, tinteger, tfloat, tboolean, tlist, tdict, required, nullable, \
    default, excludes, schema, allowed, stlist, stdict


import itertools
import re
import os
import typing as typ
from collections import namedtuple
from copy import deepcopy

import cytoolz as tz
import yaml
from funcy import *
from toposort import toposort

from quinine.common.utils import get_only_paths, allequal
from quinine.quinfig import Quinfig

# Parameter = namedtuple('Parameter', 'path dotpath value')
# SweptParameter = namedtuple('SweptParameter', 'path sweep')
SweptDisjointParameter = namedtuple('SweptDisjointParameter', 'path disjoint')
SweptProductParameter = namedtuple('SweptProductParameter', 'path product')
SweptDefaultParameter = namedtuple('SweptDefaultParameter', 'path default')


class Parameter:

    def __init__(self, path=None, dotpath=None, value=None):
        self.path = path
        self.dotpath = dotpath
        self.value = value

    def __eq__(self, other):
        if hasattr(other, 'dotpath'):
            return self.dotpath == other.dotpath
        else:
            return self.dotpath == other

    def __repr__(self):
        return f'Parameter(path={self.path}, dotpath={self.dotpath}, value={self.value})'


class SweptParameter:

    def __init__(self, path=None, sweep=None):
        self.path = path
        self.sweep = sweep

    def __getitem__(self, item):
        if item == 0:
            return self.path
        elif item == 1:
            return self.sweep

    def __repr__(self):
        return f'SweptParameter(path={self.path}, sweep={self.sweep})'


class QuinSweep:
    """
    The QuinSweep class can be used to
    - process a parameter sweep, with support for complex conditional sweeps
    - generate Quinfigs for each setting of the parameter sweep

    e.g.

    """
    # inside any parameter, specify a sweep with the special ~ character
    SWEEP_PREFIX = '~'

    # 'default' *must* be the last entry in this list
    SWEEP_TOKENS = ['disjoint', 'product'] + ['default']

    @staticmethod
    def get_parameter_path(token_path: typ.List[str]) -> typ.List[str]:
        """
        Given the path to a sweep token
        e.g.
        ['model', 'architecture', 'n_layers', '~product']
        which points to the sweep token '~product' for the 'n_layers' parameter

        get_parameter_path returns the path to the parameter on which the sweep token was applied,
        i.e.
        ['model', 'architecture', 'n_layers'].
        """
        return compose(tuple, butlast)(token_path)

    @staticmethod
    def parse_sweep_config(sweep_config) -> typ.List[typ.List[str]]:
        """
        Takes in a sweep config and determines paths to the sweep tokens in the config.

        Uses the fact that all sweep tokens in the config are prefixed by the SWEEP_PREFIX (~) character.

        Each token path
        """
        # To identify the swept parameters,
        # use the fact that any parameter sweep must use the special prefix (~ by default)
        token_paths = get_only_paths(sweep_config,
                                     pred=lambda p: any(lambda s: QuinSweep.SWEEP_PREFIX in str(s), p),
                                     stop_at=QuinSweep.SWEEP_PREFIX)

        # As output, we get a list of paths that point to locations of all the ~ tokens in the config
        token_paths = list(map(tuple, token_paths))

        # Confirm that all the tokens followed by ~ are correctly specified
        all_tokens_ok = all(lambda s: s.startswith(QuinSweep.SWEEP_PREFIX) and
                                      s.strip(QuinSweep.SWEEP_PREFIX) in QuinSweep.SWEEP_TOKENS,
                            map(last, token_paths)
                            )
        assert all_tokens_ok, f'Unknown token: sweep config failed parsing. ' \
                              f'Only tokens {QuinSweep.SWEEP_TOKENS} are allowed.'

        return token_paths

    @staticmethod
    def parse_fixed_parameters(sweep_config):
        # Extract the locations of the other non-swept, fixed parameters
        # use the fact that the other parameters must not be prefixed by the special prefix
        fixed_parameters = get_only_paths(sweep_config,
                                          pred=lambda p: all(lambda s: QuinSweep.SWEEP_PREFIX not in str(s), p))

        # Make Parameter objects
        fixed_parameters = list(map(lambda tp: Parameter(tp,
                                                         f'{".".join(map(str, tp))}.0',
                                                         get_in(sweep_config, tp)),
                                    fixed_parameters)
                                )
        return fixed_parameters

    @staticmethod
    def path_to_dotpath(path: typ.List) -> str:
        """
        Takes in any path
        e.g.
        ['model', 'architecture', 'n_layers']
        and converts it to a dotpath
        i.e.
        'model.architecture.n_layers'.
        """
        return ".".join(path)

    @staticmethod
    def dotpath_to_path(dotpath):
        """
        Takes in a dotpath
        e.g.
        'model.architecture.n_layers'
        and converts it to a path
        e.g.
        ['model', 'architecture', 'n_layers'].

        Only gin parameters are allowed to have dots in their name, and
        this fn correctly handles conversion for those cases:
        e.g.
        taking 'model.gin.architecture.n_layers' and converting it into ['model', 'gin', 'architecture.n_layers'].
        """
        # Split up the dotpath
        split = dotpath.split(".")

        # Handle the case where the gin parameter has a dot
        if len(split) > 2 and split[-3] == 'gin':
            split = split[:-2] + [".".join(split[-2:])]

        return split

    @staticmethod
    def parse_ref_dotpath(dotpath):
        """
        Reference dotpaths consist of this.is.a.parameter.idx.this.is.another.parameter._.parameter.idx (_ indicates any idx)

        Returns a list of (dotpath, idx) tuples.
        """
        return re.findall('((?:\w[^.]*)+(?:\.\w[^.]*)*?)\.(\d+|_)', dotpath)

    @staticmethod
    def param_comparator(p, q):
        """
        A comparison operator that checks which of two sweeps to apply first.

        The comparison relies on the conditional dependencies expressed in both sweeps.
        """
        # Create dotpaths for both sweeps
        p_dotpath = ".".join(p.path)
        q_dotpath = ".".join(q.path)

        # Get the keys that each sweep refers (these are the conditional dependencies that the sweep has)
        if isinstance(p, SweptDisjointParameter):
            p_ref_dotpaths = p.disjoint.keys()
        elif isinstance(p, SweptProductParameter):
            p_ref_dotpaths = p.product.keys()
        else:
            raise NotImplementedError
        if isinstance(q, SweptDisjointParameter):
            q_ref_dotpaths = q.disjoint.keys()
        elif isinstance(q, SweptProductParameter):
            q_ref_dotpaths = q.product.keys()
        else:
            raise NotImplementedError

        # Each key is a combination of dotpaths and index references: parse them to extract the dotpaths
        p_ref_parsed_dotpaths = list(
            mapcat(compose(list, autocurry(map)(0), QuinSweep.parse_ref_dotpath), p_ref_dotpaths))
        q_ref_parsed_dotpaths = list(
            mapcat(compose(list, autocurry(map)(0), QuinSweep.parse_ref_dotpath), q_ref_dotpaths))

        # Check if any reference dotpath in q matches p or vice-versa
        if any(lambda e: p_dotpath == e or p_dotpath.endswith(f'.{e}'), q_ref_parsed_dotpaths):
            return 1
        elif any(lambda e: q_dotpath == e or q_dotpath.endswith(f'.{e}'), p_ref_parsed_dotpaths):
            return -1
        else:
            return 0

    @staticmethod
    def is_product_subtype(subtype):
        return 'SweptProductParameter' in subtype.__name__

    @staticmethod
    def is_disjoint_subtype(subtype):
        return 'SweptDisjointParameter' in subtype.__name__

    def expand_partial_dotpath(self, partial_dotpath):
        """
        Given a partial dotpath, expand the dotpath to yield a full dotpath from the root to the parameter.
        """
        # This assertion checks if the reference made by the partial dotpath points to another parameter *uniquely*.
        # If not, it's impossible to know which parameter was being referenced.
        assert one(lambda k: k.endswith(partial_dotpath),
                   self.swept_parameters_dict.keys())  # swept_parameter_dict.keys() -> all_dotpaths

        # Complete the dotpath, yielding the full dotpath that points to the location of the parameter
        return list(filter(lambda t: t[1] is True,
                           map(lambda k: (k, k.endswith(partial_dotpath)),
                               self.swept_parameters_dict.keys())
                           )
                    )[0][0]

    def expand_reference(self, reference):
        # Parse the reference dotpath
        parsed_ref_dotpath = self.parse_ref_dotpath(reference)

        # Map to expand the dotpaths in the reference
        parsed_ref_dotpath = list(map(lambda t: (self.expand_partial_dotpath(t[0]), t[1]), parsed_ref_dotpath))

        # Join back
        return ".".join(list(cat(parsed_ref_dotpath)))

    def __init__(self,
                 sweep_config_path=None,
                 schema_path=None,
                 sweep_config=None,
                 schema=None
                 ):

        # Load up the sweep config
        if sweep_config is None:
            assert sweep_config_path is not None, 'Please pass in either sweep_config or sweep_config_path.'
            assert sweep_config_path.endswith('.yaml'), 'Must use a YAML file for the sweep_config.'
            sweep_config = yaml.load(open(sweep_config_path),
                                     Loader=yaml.FullLoader)

        # First, extract the locations of the sweeps being performed
        self.sweep_paths = QuinSweep.parse_sweep_config(sweep_config)

        if len(self.sweep_paths) == 0:
            # No sweep: just return the single Quinfig
            self.quinfigs = [Quinfig(config=sweep_config)]
            print(f"Generated {len(self.quinfigs)} quinfig(s) successfully.")
            return

        # Create list of paths to all the parameters that are being swept
        self.swept_parameter_paths = list(distinct(map(QuinSweep.get_parameter_path,
                                                       self.sweep_paths),
                                                   key=tuple)
                                          )

        # Next, extract the fixed parameters from the sweep config
        self.fixed_parameters = QuinSweep.parse_fixed_parameters(sweep_config)

        # Next, fetch the SweptParameter named tuples after creating them
        self.swept_parameters, \
        self.swept_disjoint_parameters, \
        self.swept_product_parameters, \
        self.swept_default_parameters = \
            self.fetch_swept_parameters(sweep_config, self.sweep_paths, self.swept_parameter_paths)

        # For convenience, create a lookup table for the swept parameters from their dotpaths
        self.swept_parameters_dict = dict(zip(map(lambda sp: ".".join(sp.path),
                                                  self.swept_parameters),
                                              self.swept_parameters)
                                          )

        # Expand all the dotpaths in any conditional
        self.expand_all_condition_dotpaths()

        # Filter out the unconditional sweeps and then process them
        uncond_disjoint_sweeps = list(filter(compose(is_seq, 1),
                                             self.swept_disjoint_parameters))
        uncond_product_sweeps = list(filter(compose(is_seq, 1),
                                            self.swept_product_parameters))

        self.all_combinations = self.process_unconditional_sweeps(uncond_disjoint_sweeps,
                                                                  uncond_product_sweeps)

        # Filter out the conditional sweeps and then process them
        cond_disjoint_sweeps = list(filter(compose(is_mapping, 1), self.swept_disjoint_parameters))
        cond_product_sweeps = list(filter(compose(is_mapping, 1), self.swept_product_parameters))

        self.process_conditional_sweeps(cond_disjoint_sweeps,
                                        cond_product_sweeps)

        # Generate all the Quinfigs
        self.quinfigs = []
        for combination in self.all_combinations:
            coll = deepcopy(sweep_config)
            for parameter in combination:
                coll = tz.assoc_in(coll, parameter.path, parameter.value)
            self.quinfigs.append(Quinfig(config=coll, base_path=os.path.dirname(os.path.abspath(sweep_config_path))))

        print(f"Generated {len(self.quinfigs)} quinfig(s) successfully.")

    def __getitem__(self, idx):
        """
        Implement indexing into the QuinSweep to fetch particular Quinfigs.
        """
        return self.quinfigs[idx]

    def expand_all_condition_dotpaths(self):
        """
        Expands the paths
        """

        # Function that expands condition dotpaths for SweptParameters
        expand_condition_dotpaths = lambda sp: \
            SweptParameter(sp.path,
                           walk_values(lambda sweep: iffy(is_mapping,
                                                          autocurry(walk_keys)(
                                                              self.expand_reference))(sweep),
                                       sp[1])
                           )

        # Apply the function to the list of SweptParameters
        self.swept_parameters = \
            list(
                map(expand_condition_dotpaths,
                    self.swept_parameters)
            )

        # Function that expands condition dotpaths for Swept___Parameters
        expand_condition_dotpaths = lambda subtype, sp: \
            subtype(sp.path,
                    iffy(is_mapping,
                         autocurry(walk_keys)(
                             self.expand_reference))(sp[1]),
                    )
        expand_condition_dotpaths = autocurry(expand_condition_dotpaths)

        self.swept_disjoint_parameters = \
            list(

                map(expand_condition_dotpaths(SweptDisjointParameter),
                    self.swept_disjoint_parameters)
            )
        self.swept_product_parameters = \
            list(
                map(expand_condition_dotpaths(SweptProductParameter),
                    self.swept_product_parameters)

            )

        # Recreate the lookup table
        self.swept_parameters_dict = dict(zip(map(lambda sp: ".".join(sp.path),
                                                  self.swept_parameters),
                                              self.swept_parameters)
                                          )

    def replace_underscores(self, swept_parameter):
        """
        Replace all the underscore references in sweep of swept_parameter.
        """

        # Find all the references (i.e. dependencies) made by the swept_parameter
        references = []
        for token in QuinSweep.SWEEP_TOKENS[:-1]:  # omit default since it's value is never a dict
            if f"~{token}" in swept_parameter.sweep and is_mapping(swept_parameter.sweep[f"~{token}"]):
                references.extend(list(swept_parameter.sweep[f"~{token}"].keys()))

        # Find all the referred parameters
        parsed_references = list(map(QuinSweep.parse_ref_dotpath, references))
        dotpaths = list(cat(parsed_references))
        ref_dict = merge_with(compose(list, cat), *list(map(lambda e: dict([e]), dotpaths)))

        # TODO: there's a bug here potentially
        assert all(map(lambda l: len(l) == len(set(l)), list(itervalues(ref_dict)))), \
            'All conditions must be distinct.'

        ref_dict_no_underscores = walk_values(compose(set,
                                                      autocurry(map)(int),
                                                      autocurry(filter)(lambda e: e != '_')),
                                              ref_dict)

        if not references:
            return swept_parameter

        def compute_possibilities(full_dotpath, reference):
            # Look up the parameter using the dotpath
            parameter = self.swept_parameters_dict[full_dotpath]

            # Use the reference to figure out how many possiblities exist for the underscore
            if len(reference) > 0:
                # Merge all the sweeps performed for this parameter
                merged_sweep = merge(*list(filter(is_mapping, itervalues(parameter.sweep))))
                # Look up the reference
                return len(merged_sweep[reference])

            assert len(parameter.sweep) == 1, 'If no reference, must be a single unconditional sweep.'
            # The number of possibilities is simply the number of values specified
            # in the (product/disjoint) unconditional sweep
            return len(list(parameter.sweep.values())[0])

        # Update the sweep by replacing underscores
        updated_sweep = swept_parameter.sweep

        # Loop over all the parsed references
        for parsed_ref in parsed_references:

            # Expand all the partial dotpaths
            # TODO: remove? expanding all the partial dotpaths in the beginning?
            parsed_ref = list(
                map(lambda t: (self.expand_partial_dotpath(t[0]), t[1]), parsed_ref))

            # For each parsed reference, there will be multiple (dotpath, idx) pairs
            for i, (full_dotpath, ref_idx) in enumerate(parsed_ref):

                # If the reference index is not an underscore, continue
                if not ref_idx == '_':
                    continue

                # Compute the prefix reference
                prefix_reference = ".".join(list(cat(parsed_ref[:i])))

                # Compute the number of possible ways to replace the underscore
                n_possibilities = compute_possibilities(full_dotpath, prefix_reference)
                replacements = set(range(n_possibilities)) - ref_dict_no_underscores[full_dotpath]

                # Find the path to the underscore condition
                path_to_condition = get_only_paths(updated_sweep, lambda p: any(lambda e: '_' in e, p),
                                                   stop_at=full_dotpath)[0]

                # Find the value of the underscore condition
                value = tz.get_in(
                    path_to_condition,
                    updated_sweep)

                # Construct keys that are subtitutes for the underscore
                keys = list(map(lambda s: f'{full_dotpath}.{s}', replacements))
                keys = list(map(lambda k: path_to_condition[:-1] + [k], keys))

                # Update by adding those keys in
                for k in keys:
                    updated_sweep = tz.assoc_in(updated_sweep, k, value)

        # Create a new swept parameter with the updated sweep
        swept_parameter = SweptParameter(swept_parameter.path,
                                         walk_values(
                                             iffy(is_mapping,
                                                  autocurry(select_keys)(lambda k: '_' not in k)
                                                  ),
                                             updated_sweep)
                                         )
        return swept_parameter

    def process_conditional_sweeps(self,
                                   cond_disjoint_sweeps,
                                   cond_product_sweeps,
                                   ):
        """
        Function to process conditional sweeps: these sweeps are applied when a conditional is satisfied.

        As an example, consider

        """
        # The complete matrix of all pairwise comparisons over the set of conditional sweeps:
        # Note that our comparison op isn't transitive (e.g. A = B and B = C but A > C is possible)
        # so we want a topological sort over the set of sweeps: O(n^2) is incurred to generate the DAG
        # over which the toposort is applied
        all_comparisons = list(map(lambda t: (t, QuinSweep.param_comparator(*t)),
                                   itertools.product(cond_product_sweeps + cond_disjoint_sweeps,
                                                     cond_product_sweeps + cond_disjoint_sweeps)))

        dependencies = merge(
            # first pretend that there are no dependencies
            dict(map(lambda t: ((t.path, type(t)), set()), cond_disjoint_sweeps)),
            dict(map(lambda t: ((t.path, type(t)), set()), cond_product_sweeps)),
            # then merge in the dependencies
            dict(map(lambda t: ((t[0][0].path, type(t[0][0])), set([(t[0][1].path, type(t[0][1])), ])),
                     filter(lambda t: t[1] == -1, all_comparisons)
                     )
                 )
        )

        # Topological sort to produce the partial ordering: a list of sets for the partial ordering
        sweep_posets = list(toposort(dependencies))

        # Map to extract the dotpaths of the sweeps
        dotpath_posets = list(map(lambda s: set(map(0, s)), sweep_posets))

        # A dotpath could occur in more than one poset
        # e.g. when it contains both a product sweep and a disjoint sweep with different conditionals
        # Ensure that the dotpath occurs exactly once, in its earliest location
        dotpath_poset_subs = list(
            reversed(list(accumulate([set()] + list(reversed(dotpath_posets))[:-1], lambda a, b: a.union(b)))))
        dotpath_posets = list(map(lambda t: t[0] - t[1], zip(dotpath_posets, dotpath_poset_subs)))

        # Loop over all the sets in the partial order
        for poset in dotpath_posets:
            # For each path, convert it to a dotpath and then replace any underscores in it
            for path in poset:
                dotpath = QuinSweep.path_to_dotpath(path)
                self.swept_parameters_dict[dotpath] = self.replace_underscores(self.swept_parameters_dict[dotpath])

        for poset in sweep_posets:
            # Split the poset into the disjoint and product sweeps
            disjoint_poset = [p for p in poset if QuinSweep.is_disjoint_subtype(p[1])]
            product_poset = [p for p in poset if QuinSweep.is_product_subtype(p[1])]

            # Process the product sweeps first
            for path, subtype in product_poset:
                dotpath = QuinSweep.path_to_dotpath(path)
                # Each parameter combination needs to be expanded
                # print("Product Expansion")
                # for thing in self.all_combinations:
                #     print([e.dotpath for e in thing])
                self.all_combinations = self.product_expansion(self.all_combinations,
                                                               dotpath,
                                                               self.swept_parameters_dict[dotpath].sweep['~product'])

            # Process the poset
            # For disjoint, process the entire poset together (restricted to the disjoint sweeps)
            self.all_combinations = self.disjoint_expansion(self.all_combinations, disjoint_poset)

            pass

        # Apply the defaults
        for poset in dotpath_posets:
            # For each path, convert it to a dotpath and then replace any underscores in it
            for path in poset:
                for i, combo in enumerate(self.all_combinations):
                    dotpath = QuinSweep.path_to_dotpath(path)
                    if dotpath not in list(map(lambda c: self.path_to_dotpath(c.path), combo)):
                        combo.append(Parameter(path,
                                               f'{dotpath}.0',
                                               self.swept_parameters_dict[dotpath].sweep['~default']))
                        self.all_combinations[i] = combo

    def product_expansion(self,
                          combinations,
                          dotpath,
                          sweep):
        """
        Takes as input

        """
        new_combinations = []
        # Iterate over the parameter combinations
        for i, combo in enumerate(combinations):
            # print(i, combo, sweep)
            flag = False
            # The parameter combination contains a setting of the previously configured parameters
            for key in sweep:
                # First parse the key, which is a reference dotpath, to a list of dotpaths
                # Check if all the ref dotpaths are satisfied by the combo
                if all(map(lambda t: ".".join(t) in combo, self.parse_ref_dotpath(key))):
                    # Perform the product expansion
                    new_combos = list(itertools.product(*list(map(lambda e: [e], combo)),
                                                        list(
                                                            map(lambda e: Parameter(self.dotpath_to_path(dotpath),
                                                                                    f'{dotpath}.{e[0]}',
                                                                                    e[1]),
                                                                enumerate(sweep[key])
                                                                )
                                                        ))
                                      )
                    new_combos = list(map(compose(list, concat), new_combos))
                    new_combinations.extend(new_combos)
                    flag = True
                    break
            if not flag:
                new_combinations.append(combo)

        return new_combinations

    def disjoint_expansion(self,
                           combinations,
                           poset):
        """

        """

        ref_dotpath_to_dotpath = {}
        for path, _ in poset:
            # Look up the SweptParameter using the path
            swept_param = self.swept_parameters_dict[self.path_to_dotpath(path)]
            # Loop over the conditionals in the SweptParameter to create the lookup table
            for ref_dotpath in swept_param.sweep['~disjoint']:
                if ref_dotpath not in ref_dotpath_to_dotpath:
                    ref_dotpath_to_dotpath[ref_dotpath] = set()
                ref_dotpath_to_dotpath[ref_dotpath].add(swept_param)

        new_combinations = []

        # Iterate over the parameter combinations
        for i, combo in enumerate(combinations):
            # print(i, combo)
            flag = False
            # The parameter combination contains a setting of the previously configured parameters
            for ref_dotpath in ref_dotpath_to_dotpath:
                # First parse the key, which is a reference dotpath, to a list of dotpaths
                # Check if all the ref dotpaths are satisfied by the combo
                if all(map(lambda t: ".".join(t) in combo, self.parse_ref_dotpath(ref_dotpath))):
                    disjoint_parameters = list(zip(*map(lambda sp: list(map(lambda v: Parameter(sp.path,
                                                                                                f'{self.path_to_dotpath(sp.path)}.{v[0]}',
                                                                                                v[1]),
                                                                            enumerate(
                                                                                sp.sweep['~disjoint'][ref_dotpath])
                                                                            )
                                                                        ),
                                                        ref_dotpath_to_dotpath[ref_dotpath])
                                                   ))

                    new_combos = list(map(compose(list, cat), zip([combo] * len(disjoint_parameters),
                                                                  disjoint_parameters)
                                          )
                                      )
                    new_combinations.extend(new_combos)
                    flag = True
                    break

            if not flag:
                new_combinations.append(combo)

        return new_combinations

    def process_unconditional_sweeps(self,
                                     uncond_disjoint_sweeps,
                                     uncond_product_sweeps):
        """
        Parameters with unconditional sweeps are always processed first, and unconditional sweeps specify how to vary
        these parameters.

        Sweeps should be thought of as trees, with the root node containing assignments to all unswept parameters.
        Each time one or a group of swept parameters are processed, the tree grows in depth. The path from the root
        of the tree to any leaf indicates a particular assignment to the swept parameters.

        Unconditional sweeps can be

        - disjoint over r parameters, each parameter taking exactly k parameter values

        Unconditional disjoint sweeps can be thought of as generating k disjoint sweeps,
        where each sweep contains assignments to the r parameters.
        The k parameter values specified for each of the r parameters are assumed to be in alignment.

        In terms of the tree structure, this operation adds k children to the root node,
         with each child containing a simultaneous assignment to the r parameters.

        - product over p parameters, with the parameters taking k_1, k_2, ..., k_p parameter values

        Unconditional product sweeps generate k_i new sweeps for the ith (of p) parameter. Each of these product sweeps
        add a new layer to the tree, leading to a total depth increase of p. The ith such layer contains assignments
        to the ith parameter.

        In total, unconditional sweeps generate a total of k * k_1 * k_2 ... * k_r total leaf nodes in the tree.
        If the sweep only contains unconditional sweeps, the leaf nodes will contain complete assignments to all
        parameters, and can be used to create a Quinfig.
        """

        # All of the disjoint unconditional sweeeps must be identical length
        assert allequal(
            map(compose(len, 1), uncond_disjoint_sweeps)), 'All disjoint unconditional sweeps must have same length.'
        disjoint_uncond_combinations = list(zip(
            *map(
                lambda t: list(
                    map(lambda e: Parameter(t.path, f'{".".join(t.path)}.{e[0]}', e[1]), enumerate(t.disjoint))),
                uncond_disjoint_sweeps))
        )
        all_combinations = disjoint_uncond_combinations

        # Next, process the sequential product sweeps
        product_uncond_combinations = list(
            itertools.product(
                *map(lambda t: list(
                    map(lambda e: Parameter(t.path,
                                            f'{".".join(t.path)}.{e[0]}',
                                            e[1]),
                        enumerate(t.product)
                        )
                ),
                     uncond_product_sweeps)
            )
        )
        all_combinations = list(
            map(compose(list, cat),
                itertools.product(all_combinations,
                                  product_uncond_combinations)
                )
        )

        return all_combinations

    def fetch_swept_parameters(self,
                               config,
                               sweep_paths,
                               swept_parameter_paths):
        """
        Construct named tuples for all the parameters being swept.
        """
        # Filter sweep paths to all 3 sweep types
        disjoint_sweep_paths = list(filter(lambda s: 'disjoint' in last(s), sweep_paths))
        product_sweep_paths = list(filter(lambda s: 'product' in last(s), sweep_paths))
        default_sweep_paths = list(filter(lambda s: 'default' in last(s), sweep_paths))

        # Construct SweptParameter and Swept__Parameter namedtuples,
        # making it
        # consisting of the path to the parameter and its sweep configuration
        construct_swept_parameters = lambda subtype, paths, wrapper: list(
            map(lambda p: subtype(wrapper(p),
                                  tz.get_in(p, coll=config)
                                  ),
                paths)
        )

        swept_parameters = construct_swept_parameters(SweptParameter,
                                                      swept_parameter_paths,
                                                      identity)
        swept_disjoint_parameters = construct_swept_parameters(SweptDisjointParameter,
                                                               disjoint_sweep_paths,
                                                               QuinSweep.get_parameter_path)
        swept_product_parameters = construct_swept_parameters(SweptProductParameter,
                                                              product_sweep_paths,
                                                              QuinSweep.get_parameter_path)
        swept_default_parameters = construct_swept_parameters(SweptDefaultParameter,
                                                              default_sweep_paths,
                                                              QuinSweep.get_parameter_path)

        return swept_parameters, swept_disjoint_parameters, swept_product_parameters, swept_default_parameters


if __name__ == '__main__':
    import os

    print(os.getcwd())
    # sweep_config = Quinfig(
    #     config_path='quinine/tests/derived-1-2.yaml')
    sweep_config = Quinfig(
        config_path='tests/bugs/test-2.yaml')
    # sweep_config = Quinfig(
    #     config_path='/Users/krandiash/Desktop/workspace/projects/quinine/tests/derived-2.yaml')

    quin_sweep = QuinSweep(sweep_config=sweep_config)
    print(quin_sweep.quinfigs)



"""
Defines the main Quinfig class.
"""

from quinine.common.cerberus import *
from quinine.common.gin import *
from quinine.common.utils import *

# Register all schemas
create_and_register_schemas()


class Quinfig(Munch):
    """
    The Quinfig class for creating configuration files.

    Quinfig provides a simple abstraction for configuration files with several useful features:
    - write (arbitrarily nested) configuration files in YAML or just specify them as dictionaries in Python
    - access keys with dot access
    - validate your configs against schemas using Cerberus
    - set fn arguments using gin
    """

    def __init__(self,
                 config_path=None,
                 schema_path=None,
                 config=None,
                 schema=None,
                 base_path=None,
                 ):
        # Prepare the config
        config = prepare_config(config_path=config_path,
                                schema_path=schema_path,
                                config=config,
                                schema=schema,
                                base_path=base_path)

        # Create the Quinfig
        super(Quinfig, self).__init__(config)

    def __repr__(self):
        """
        Use a yaml dump to create a formatted representation for the Quinfig.
        """
        return f'Quinfig\n' \
               f'-------\n' \
               f'{yaml.dump(self.__dict__)}'


def prepare_config(config_path=None,
                   schema_path=None,
                   config=None,
                   schema=None,
                   base_path=None) -> Munch:
    """
    Takes in paths to config and schema files.
    Validates the config against the schema, normalizes the config, parses gin and converts the config to a Munch.
    """
    # Load up the config
    if config is None:
        assert config_path is not None, 'Please pass in either config or config_path.'
        assert config_path.endswith('.yaml'), 'Must use a YAML file for the config.'
        config = yaml.load(open(config_path),
                           Loader=yaml.FullLoader)

    # If the config is a Quinfig object, just grab the __dict__ for convenience
    if isinstance(config, Quinfig):
        config = config.__dict__

    # Convert config to Munch: iffy ensures that the Munch fn is only applied to mappings
    config = walk_values_rec(iffy(is_mapping, lambda c: Munch(**c)), config)

    # Load up the schema
    if schema is None:
        if schema_path is not None:
            assert schema_path.endswith('.yaml'), 'Must use a YAML file for the config.'
            schema = yaml.load(open(schema_path),
                               Loader=yaml.FullLoader)

    if schema is not None:
        # Allow gin configuration at any level of nesting: put a gin tag at every level of the schema
        schema = autoexpand_schema(schema)

        # Validate the config against the schema
        validate_config(config, schema)

    # Normalize the config
    if not base_path:
        base_path = os.path.dirname(os.path.abspath(config_path)) if config_path else ''
    else:
        base_path = os.path.abspath(base_path)
    config = normalize_config(config, schema, base_path=base_path)

    # Parse and load the gin configuration
    nested_gin_dict_parser(config)

    return config



import gin
from funcy import *
from quinine.common.utils import nested_dict_walker, prefix


def register_module_with_gin(module, module_name=None):
    """
    Register all the callables in a single module with gin.

    A useful way to add gin configurability to a codebase without explicilty using the @gin.configurable decorator.
    """
    module_name = module.__name__ if module_name is None else module_name

    for attr in dir(module):
        if callable(getattr(module, attr)):
            setattr(module, attr, gin.configurable(getattr(module, attr), module=module_name))


def scope_datagroup_gin_dict(coll):
    """
    Rename the augmentations gin dict with the alias of the dataset.
    """

    if 'alias' in coll and 'gin' in coll and is_mapping(coll['gin']):
        coll['gin'] = walk_keys(prefix(p=f"{coll['alias'].replace('.', '_')}/"), coll['gin'])
    return coll


def nested_scope_datagroup_gin_dict(coll):
    """
    Apply the renamer over a nested dict, e.g. derived from a yaml.
    """
    return nested_dict_walker(scope_datagroup_gin_dict, coll)


def gin_dict_parser(coll):
    """
    Use for parsing collections that may contain a 'gin' key.
    The 'gin' key is assumed to map to either a dict or str value that contains gin bindings.
    e.g.
    {'gin': {'Classifier.n_layers': 2, 'Classifier.width': 3}}
    or
    {'gin': 'Classifier.n_layers = 2\nClassifier.width = 3'}
    """
    if 'gin' in coll:
        if is_mapping(coll['gin']):
            gin.parse_config("".join(map(lambda t: f'{t[0]} = {t[1]}\n', iteritems(coll['gin']))))
        elif isinstance(coll['gin'], str):
            gin.parse_config(coll['gin'])
    return coll


def nested_gin_dict_parser(coll):
    """
    Use for parsing nested collections that may contain a 'gin' key.
    The 'gin' key is assumed to map to a dict value that contains gin bindings (see gin_dict_parser).

    Enables support for gin keys in yaml files.
    """
    return nested_dict_walker(gin_dict_parser, coll)



import sys as _sys
from argparse import ArgumentParser

import cytoolz as tz
from funcy import *

from quinine.common.cerberus import tstring, tlist
from quinine.common.utils import rmerge
from quinine.quinfig import Quinfig
from quinine.quinfig import get_all_leaf_paths


class QuinineArgumentParser(ArgumentParser):
    """
    Class for replacing the standard argparse.ArgumentParser.

    In addition to the standard functionality of ArgumentParser,
    - includes an argument for '--config', which passes in a YAML file containing the configuration parameters
    - automatically adds arguments to the ArgumentParser from a provided schema,
    allowing you to override arguments in your YAML directly from the command line
    """
    types = {'string': str,
             'integer': int,
             'float': float,
             'dict': dict,
             'boolean': bool,
             'list': list,
             }

    def __init__(self, schema=None, **kwargs):
        super(QuinineArgumentParser, self).__init__(**kwargs)

        # Add a default argument for the path to the YAML configuration file that will be passed in
        self.add_argument('--config', type=str, required=True, help="YAML configuration file.")
        self.schema = schema

        if self.schema is not None:
            # Populate the argument parser with arguments from the schema
            paths_to_type = list(filter(lambda l: l[-1] == 'type', get_all_leaf_paths(self.schema)))
            type_lookup = dict([(tuple(filter(lambda e: e != 'schema', e[:-1])), tz.get_in(e, schema))
                                for e in paths_to_type])

            valid_params = self.get_all_params(schema)
            for param in valid_params:
                self.add_argument(f'--{".".join(param)}',
                                  type=self.types[type_lookup[param]])

        self.schema = merge(self.schema, {'config': tstring, 'inherit': tlist})

    @staticmethod
    def get_all_params(schema):
        # Find all leaf paths in the schema, then truncate the last key from each path
        # and remove the 'schema' key if it occurs anywhere in the path
        # TODO: expand the list of criteria in the inner lambda
        candidate_parameters = list(set(map(lambda l: tuple(filter(lambda e: e != 'schema' and e != 'allowed', l[:-1])),
                                            get_all_leaf_paths(schema))
                                        )
                                    )

        # Remove prefix paths from the candidate parameters,
        # e.g. when ['general', 'seed'] and ['general'] both occur, remove ['general']
        valid_parameters = set()
        all_subpaths = set()
        for path in sorted(candidate_parameters, key=lambda l: len(l), reverse=True):
            # If the path isn't in the set of subpaths seen so far, it's a valid param (because paths are sorted)
            if path not in all_subpaths:
                valid_parameters.add(path)
                # Add all subpaths in for this
                for i in range(1, len(path)):
                    all_subpaths.add(path[:i])

        return valid_parameters

    def parse_quinfig(self):
        # Parse all the arguments from the command line, overriding defaults in the argparse
        args = self.parse_args()

        cli_keys = []
        for cli_arg in _sys.argv[1:]:
            if cli_arg.startswith('--'):
                if str(cli_arg) == '--config':
                    continue
                cli_keys.append(cli_arg[2:].replace("-", "_"))
            elif cli_arg.startswith('-'):
                raise NotImplementedError("QuinineArgumentParser doesn't support abbreviated arguments.")
            else:
                continue

        # Get parameters which need to be overridden from command line
        override_args = project(args.__dict__, cli_keys)

        # Trick: first load the config without a schema as a base config
        # quinfig = Quinfig(config_path=args.config)

        # Override all the defaults using the yaml config
        # quinfig = rmerge(Quinfig(config=args.__dict__), quinfig)
        # print(quinfig)

        # Use all the defaults in args to populate a dictionary
        quinfig = {}
        for param, val in args.__dict__.items():
            param_path = param.split(".")
            quinfig = tz.assoc_in(quinfig, param_path, val)

        # Override all the defaults using the yaml config
        quinfig = rmerge(quinfig, Quinfig(config_path=args.config))

        # Replace all the arguments passed into command line
        if len(override_args) > 0:
            print(f"\n(quinine) Overriding parameters in {args.config} from command line (___ is unspecified).")

        for param, val in override_args.items():
            param_path = param.split(".")
            old_val = tz.get_in(param_path, quinfig)

            if old_val != val:
                print(f"> ({param}): {old_val} --> {val}")
            else:
                print(f"> ({param}): ___ --> {val}")
            quinfig = tz.assoc_in(quinfig, param_path, val)

        print()
        # Load the config again
        return Quinfig(config=dict(quinfig), schema=self.schema)



"""
Some common utilities.
"""

import yaml
from funcy import *
from munch import Munch
import cytoolz as tz


def difference(*colls):
    """
    Find the keys that have different values in an arbitrary number of (nested) collections. Any key
    that differs in at least 2 collections is considered to fit this criterion.
    """

    # Get all the leaf paths for each collection: make each path a tuple
    leaf_paths_by_coll = list(map(lambda c: list(map(tuple, get_all_leaf_paths(c))), colls))

    # Find the union of all leaf paths: merge all the paths and keep only the unique paths
    union_leaf_paths = list(distinct(concat(*leaf_paths_by_coll)))

    # Get the values corresponding to these leaf paths in every collection: if a leaf path doesn't exist, assumes None
    values_by_coll = list(map(lambda lp: list(map(lambda coll: tz.get_in(lp, coll), colls)), union_leaf_paths))

    # Filter out the leaf paths that have identical values across the collections
    keep_leaf_paths = list(map(0, filter(lambda t: not allequal(t[1]), zip(union_leaf_paths, values_by_coll))))
    keep_values = list(map(1, filter(lambda t: not allequal(t[1]), zip(union_leaf_paths, values_by_coll))))

    # Rearrange to construct a list of dictionaries -- one per original collection.
    # Each of these dictionaries maps a 'kept' leaf path to its corresponding
    # value in the collection
    differences = list(map(lambda vals: dict(zip(keep_leaf_paths, vals)), list(zip(*keep_values))))

    return differences



def rmerge(*colls):
    """
    Recursively merge an arbitrary number of collections.
    For conflicting values, later collections to the right are given priority.
    Note that this function treats sequences as a normal value and sequences are not merged.

    Uses:
    - merging config files
    """
    if isinstance(colls, tuple) and len(colls) == 1:
        # A squeeze operation since merge_with generates tuple(list_of_objs,)
        colls = colls[0]
    if all(is_mapping, colls):
        # Merges all the collections, recursively applies merging to the combined values
        return merge_with(rmerge, *colls)
    else:
        # If colls does not contain mappings, simply pick the last one
        return last(colls)


def prettyprint(s):
    if hasattr(s, '__dict__'):
        print(yaml.dump(s.__dict__))
    elif isinstance(s, dict):
        print(yaml.dump(s))
    else:
        print(s)


def allequal(seq):
    return len(set(seq)) <= 1


@autocurry
def listmap(fn, seq):
    return list(map(fn, seq))


@autocurry
def prefix(s, p):
    if isinstance(s, str):
        return f'{p}{s}'
    elif isinstance(s, list):
        return list(map(prefix(p=p), s))
    else:
        raise NotImplementedError


@autocurry
def postfix(s, p):
    if isinstance(s, str):
        return f'{s}{p}'
    elif isinstance(s, list):
        return list(map(postfix(p=p), s))
    else:
        raise NotImplementedError


@autocurry
def surround(s, pre, post):
    return postfix(prefix(s, pre), post)


def nested_map(f, *args):
    """ Recursively transpose a nested structure of tuples, lists, and dicts """
    assert len(args) > 0, 'Must have at least one argument.'

    arg = args[0]
    if isinstance(arg, tuple) or isinstance(arg, list):
        return [nested_map(f, *a) for a in zip(*args)]
    elif isinstance(arg, dict):
        return {
            k: nested_map(f, *[a[k] for a in args])
            for k in arg
        }
    else:
        return f(*args)


@autocurry
def walk_values_rec(f, coll):
    """
    Similar to funcy's walk_values, but does so recursively, including mapping f over lists.
    """
    if is_mapping(coll):
        return f(walk_values(walk_values_rec(f), coll))
    elif is_list(coll):
        return f(list(map(walk_values_rec(f), coll)))
    else:
        return f(coll)


@autocurry
def nested_dict_walker(fn, coll):
    """
    Apply a function over the mappings contained in coll.
    """
    return walk_values_rec(iffy(is_mapping, fn), coll)


def get_all_leaf_paths(coll):
    """
    Returns a list of paths to all leaf nodes in a nested dict.
    Paths can travel through lists and the index is inserted into the path.
    """
    if isinstance(coll, dict) or isinstance(coll, Munch):
        return list(cat(map(lambda t: list(map(lambda p: [t[0]] + p,
                                               get_all_leaf_paths(t[1])
                                               )),
                            iteritems(coll)))
                    )

    elif isinstance(coll, list):
        return list(cat(map(lambda t: list(map(lambda p: [t[0]] + p,
                                               get_all_leaf_paths(t[1])
                                               )),
                            enumerate(coll)))
                    )
    else:
        return [[]]


def get_all_paths(coll, prefix_path=(), stop_at=None, stop_below=None):
    """
    Given a collection, by default returns paths to all the leaf nodes.
    Use stop_at to truncate paths at the given key.
    Use stop_below to truncate paths one level below the given key.
    """
    assert stop_at is None or stop_below is None, 'Only one of stop_at or stop_below can be used.'
    if stop_below is not None and stop_below in str(last(butlast(prefix_path))):
        return [[]]
    if stop_at is not None and stop_at in str(last(prefix_path)):
        return [[]]
    if isinstance(coll, dict) or isinstance(coll, Munch) or isinstance(coll, list):
        if isinstance(coll, dict) or isinstance(coll, Munch):
            items = iteritems(coll)
        else:
            items = enumerate(coll)

        return list(cat(map(lambda t: list(map(lambda p: [t[0]] + p,
                                               get_all_paths(t[1],
                                                             prefix_path=list(prefix_path) + [t[0]],
                                                             stop_at=stop_at,
                                                             stop_below=stop_below)
                                               )),
                            items))
                    )
    else:
        return [[]]


def get_only_paths(coll, pred, prefix_path=(), stop_at=None, stop_below=None):
    """
    Get all paths that satisfy the predicate fn pred.
    First gets all paths and then filters them based on pred.
    """
    all_paths = get_all_paths(coll, prefix_path=prefix_path, stop_at=stop_at, stop_below=stop_below)
    return list(filter(pred, all_paths))

if __name__ == '__main__':
    coll1 = {'a': 1,
             'b': 2,
             'c': {'d': 12}}
    coll2 = {'a': 1,
             'b': 2,
             'c': {'d': 13}}
    coll3 = {'a': 1,
             'b': 3,
             'c': {'d': 14},
             'e': 4}

    difference(coll1, coll2, coll3)


import glob
import os

import cytoolz.curried as tz
from cerberus import schema_registry, Validator
from quinine.common.utils import *
from quinine.common.gin import nested_scope_datagroup_gin_dict

tstring = {'type': 'string'}
tinteger = {'type': 'integer'}
tfloat = {'type': 'float'}
tboolean = {'type': 'boolean'}
tlist = {'type': 'list'}
tlistorstring = {'type': ['list', 'string']}
tdict = {'type': 'dict'}

required = {'required': True}
nullable = {'nullable': True}
default = lambda d: {'default': d}
excludes = lambda e: {'excludes': e}
schema = lambda s: {'schema': s}
allowed = lambda a: {'allowed': a}

stlist = lambda s: merge(tlist, schema(merge(tdict, schema(s))))
stlistorstring = lambda s: merge(tlistorstring, schema(merge(tdict, schema(s))))
stdict = lambda s: merge(tdict, schema(s))


def create_and_register_schemas():
    # A schema for gin data
    gin = {'gin': tdict}

    # Loader schema for creating e.g. DataLoaders in torch
    loader = {'loader': stdict({
        'strategy': tstring,
        'batch_size': tinteger,
        'num_workers': tinteger,
        'seed': merge(tinteger, default(0)),
    })}

    # A standard block for data loaders that is used to create a hierarchical schema for dataflows
    datablock = {'source': merge(tstring, allowed(['torchvision', 'imagefolder', 'tfds', 'tfrecord'])),
                 'name': tstring,
                 'shuffle_and_split': tboolean,
                 'load_path': tstring,
                 'seed': tinteger,
                 'heads': merge(tlist, nullable, default(None)),
                 }
    datablock = merge(datablock, loader, gin)

    # Schema for a group: a group is like a slice of data from a particular dataset
    datagroup = merge(datablock,
                      {'split_cmd': merge(tstring, required),
                       'alias': merge(tstring, required),
                       })
    datagroups = {'groups': stlist(datagroup)}

    # A datasets schema for building a list of datasets, with each dataset containing multiple groups
    datasets = {'datasets': stlist(merge(datablock, datagroups))}

    # A dataflow: complete description of all the datasets and groups, their loaders as well as where they are used
    dataflow = merge(datablock,
                     {'train': stdict(merge(datablock,
                                            datasets,
                                            {'loader_level': merge(tstring,
                                                                   required,
                                                                   allowed(['all', 'dataset', 'group']))})),
                      'val': stdict(merge(datablock, datasets,
                                          {'loader_level': merge(tstring,
                                                                 default('group'),
                                                                 allowed(['all', 'dataset', 'group']))})),
                      'test': stdict(merge(datablock, datasets,
                                           {'loader_level': merge(tstring,
                                                                  default('group'),
                                                                  allowed(['all', 'dataset', 'group']))}))
                      })

    # Schema for the duration of training
    duration = {'epochs': merge(tinteger, required, excludes('steps')),
                'steps': merge(tinteger, required, excludes('epochs'))}

    # Schema for the optimizer
    optimizer = {'optimizer': stdict(merge({'name': merge(tstring, required)}, gin))}

    # Schema for the learning rate scheduler
    lr = {'lr': stdict(merge({'scheduler': merge(tstring, required)}, gin))}

    # Trainer: complete description of how to update the model
    trainer = merge(duration, optimizer, lr, gin)

    # Schema for constructing the model
    model = {'source': merge(tstring, required, allowed(['torchvision', 'torch', 'cm', 'keras'])),
             'architecture': merge(tstring, required),
             'finetuning': merge(tboolean, required),
             'pretrained': merge(tboolean, required),
             'heads': merge(tlist, required),
             }
    model = merge(model, gin)

    # Schema for the checkpointer
    checkpointer = {'ckpt_path': merge(tstring, nullable, default(None)),
                    'save_freq': merge(tinteger, required)}
    checkpointer = merge(checkpointer, gin)

    # Schema for resuming model training from Weights and Biases
    wandb_resumer = {'resume': merge(tboolean, default(False)),
                     'run_id': merge(tstring, default(None), nullable),
                     'project': merge(tstring, default(None), nullable),
                     'entity': merge(tstring, default(None), nullable),
                     }
    wandb_resumer = merge(wandb_resumer, gin)

    # Schema for Weights and Biases
    wandb = {'entity': merge(tstring, required),
             'project': merge(tstring, required),
             'group': merge(tstring, default('default')),
             'job_type': merge(tstring, default('training')),
             'ckpt_dir': merge(tstring, default('checkpoints')),
             'dryrun': merge(tboolean, default(False)),
             }
    wandb = merge(wandb, gin)

    # Schema for general settings
    general = {'seed': merge(tinteger, default(0)),
               'module': merge(tstring, required),
               }

    # Schema for Kubernetes
    kubernetes = {}

    # Schema for templating
    templating = {'parent': merge(tstring, nullable, default(None))}

    # Schema for inheritance
    inherit = stlistorstring(merge(tstring, nullable, default(None)))

    # Collect all the schemas that are reusable
    schemas = {
        'general': general,

        'duration': duration,
        'optimizer': optimizer,
        'lr': lr,
        'trainer': trainer,

        'checkpointer': checkpointer,
        'wandb_resumer': wandb_resumer,

        'model': model,

        'kubernetes': kubernetes,
        'templating': templating,
        'inherit': inherit,

        'loader': loader,
        'datablock': datablock,
        'datagroup': datagroup,
        'datagroups': datagroups,
        'datasets': datasets,
        'dataflow': dataflow,

        'gin': gin,

        'wandb': wandb,
    }

    # Register the schemas
    register_schemas(*list(zip(*schemas.items())), verbose=False)

    return schemas


def register_schemas(schema_names, schemas, verbose=True):
    """
    Register a list of schemas, with corresponding names.
    """
    # Register the schemas
    list(map(lambda n, s: schema_registry.add(n, s), schema_names, schemas))

    if verbose:
        # Print
        print("Registered schemas in Cerberus: ")
        list(map(lambda n: print(f'- {n}'), schema_names))


def register_yaml_schemas(path):
    """
    Register all schemas located in a directory.
    Schemas are assumed to be defined in yaml files at path.
    """
    # Get the schema files
    schema_files = glob.glob(os.path.join(path, '*'))
    schema_names = list(map(lambda f: os.path.basename(f).replace(".yaml", ""), schema_files))
    schemas = list(map(compose(autocurry(yaml.load)(Loader=yaml.FullLoader), open), schema_files))

    # Register them
    register_schemas(schema_names, schemas)


def normalize_config(config, schema=None, base_path=''):
    """
    Execute a series of functions on the config that modify it.
    """
    if schema:
        config = Validator(schema).normalized(config)
    config = resolve_templating(config)
    config = resolve_inheritance(config, base_path=base_path)
    if 'dataflow' in config:
        dataflow_config = propagate_parameters_to_datagroups(config.dataflow)
        config = set_in(config, ['dataflow'], dataflow_config)
    config = nested_scope_datagroup_gin_dict(config)
    return config


def resolve_inheritance(config, base_path=''):
    """
    Takes in a config and resolves any inheritance.
    If inheriting, the config will have information about one or more parent configs that should be overwritten
    (those configs may in turn inherit from others).
    This inheritance chain is resolved by recursively merging all the relevant configs.
    """
    if 'inherit' not in config or ('inherit' in config and config['inherit'] is None):
        return config

    inherit_paths = [config['inherit']] if isinstance(config['inherit'], str) else config['inherit']
    inherit_paths = [os.path.abspath(os.path.join(base_path, inherit_path)) for inherit_path in inherit_paths]
    config['inherit'] = inherit_paths

    inherit_configs = [
        # Recurse to resolve inheritance for each inherited config
        resolve_inheritance(
            yaml.load(open(inherit_path),
                      Loader=yaml.FullLoader),
            base_path=os.path.dirname(inherit_path),
        )
        for inherit_path in inherit_paths
    ]

    config = rmerge(*inherit_configs, config)
    return config


def resolve_templating(config, base_path=''):
    """
    Takes in a config and resolves any templating.
    If templating, the config will have information about a parent config that it is overwriting
    (which in turn may itself be templating).
    This templating chain is resolved by recursively merging all the relevant configs.
    """
    if 'templating' not in config or ('templating' in config and config['templating']['parent_yaml'] is None):
        return config

    append_parent = lambda l: [yaml.load(open(os.path.join(base_path, l[0]['templating']['parent_yaml'])),
                                         Loader=yaml.FullLoader)] + l
    construct_hierarchy = lambda l: ignore(errors=Exception,
                                           default=append_parent(l)
                                           )(construct_hierarchy)(append_parent(l))
    config_hierarchy = construct_hierarchy([config])
    config = rmerge(*config_hierarchy)
    return config


def validate_config(config, schema):
    """
    Check if a config file adheres to a schema.
    """
    validator = Validator(schema)
    valid = validator.validate(config)
    if valid:
        return True
    else:
        print("CerberusError: config could not be validated against schema. The errors are,")
        print(validator.errors)
        exit()


def expand_schema_for_gin_configuration(schema):
    """
    Allows the schema to support gin configurability.
    The schema supports gin keys at any level of nesting.
    """
    # Insert a 'gin' key into the dictionaries that don't contain a 'type' key
    predicate = lambda p: 'type' not in p

    # Merge a gin schema into the schema passed in, at every level of nesting
    return nested_dict_walker(iffy(predicate,
                                   lambda v: merge({'gin': tdict}, v)),
                              schema)


def expand_schema_for_inheritance(schema):
    """
    Allows the schema to support inheritance.
    The schema supports configs that (optionally) point to zero or more parent YAML configs
    (using a path) that will be taken as base configurations to be overwritten.
    """
    return merge({'inherit': stlistorstring(merge(tstring,
                                                  nullable,
                                                  default(None))
                                            )
                  }, schema)


def expand_schema_for_templating(schema):
    """
    Allows the schema to support templating.
    The schema supports configs that (optionally) point to a parent YAML config (using a path) that is being overwritten.
    TODO: Deprecate.
    """
    return merge({'templating': stdict({'parent_yaml': merge(tstring,
                                                             nullable,
                                                             default(None))
                                        })
                  }, schema)


def autoexpand_schema(schema):
    """
    Automatically expands the schema to support
    - gin configuration
    - templating
    """
    schema = expand_schema_for_gin_configuration(schema)
    schema = expand_schema_for_templating(schema)
    schema = expand_schema_for_inheritance(schema)
    return schema


def propagate_parameters_to_datagroups(dataflow_config):
    """
    Given a dataflow config, it is likely that parameters were defined 'globally' e.g. setting the dataset's source
    for all datasets and groups. This function propagates parameters from higher levels down to the lowest, group level.
    The propagation consolidates parameters at the following levels:
    - dataflow
    - train(/val/test)
    - datasets[i]
    - groups[j]
    into the parameters that are applicable to groups[j].
    """

    def construct_group_dict(group_path, config):
        """
        Given a config and a path that points to a data group, compute the data group's updated parameters.
        The group_path is a list of keys and indices e.g. ['train', 'datasets', 1, 'groups', 0]
        that can be followed to reach a group's config.
        """
        # Find (almost) all prefixes of the group path
        all_paths = list(map(compose(list,
                                     tz.take(seq=group_path)),
                             range(1, len(group_path))
                             )
                         )

        # Filter to exclude paths that point to lists
        paths_to_merge = list(filter(lambda p: isinstance(last(p[1]), str),
                                     pairwise(all_paths)
                                     )
                              )
        # Find all the (mid-level) dicts that the filtered paths point to
        mid_level_dicts = list(map(lambda p: tz.keyfilter(lambda k: k != last(p[1]),
                                                          tz.get_in(p[0], config)),
                                   paths_to_merge))

        # Merge parameters at all levels to get a single parameter set for the group
        def dmerge(*args):
            if all(is_mapping, *args):
                return Munch(tz.merge(*args))
            else:
                return tz.last(*args)

        group_dict = tz.merge_with(
            dmerge,
            tz.keyfilter(lambda k: k not in ['train', 'val', 'test'], config),  # top-level dict
            *mid_level_dicts,  # mid-level dicts
            tz.get_in(group_path, config)  # bottom-level dict
        )

        return group_dict

    def get_all_group_paths(config, following=()):
        """
        Given a config, constructs paths to all the leaf nodes, truncating them one level below the 'groups' key.
        """
        if isinstance(config, dict) or isinstance(config, Munch):
            if 'groups' in list(butlast(following)):
                return [[]]
            return list(cat(map(lambda t: list(map(lambda p: [t[0]] + p,
                                                   get_all_group_paths(t[1], list(following) + [t[0]])
                                                   )),
                                iteritems(config)))
                        )

        elif isinstance(config, list):
            return list(cat(map(lambda t: list(map(lambda p: [t[0]] + p,
                                                   get_all_group_paths(t[1], list(following) + [t[0]])
                                                   )),
                                enumerate(config)))
                        )
        else:
            return [[]]

    # Find all the group paths
    group_paths = list(filter(lambda p: 'groups' in p,
                              get_all_group_paths(dataflow_config)
                              )
                       )

    # Construct the group dict for each group path
    group_dicts = list(map(lambda p: construct_group_dict(p,
                                                          dataflow_config),
                           group_paths)
                       )

    # Update the dataflow_config with all the group dicts
    updated_dataflow_config = compose(*list(map(autocurry(lambda p, d, c: set_in(c, p, d)),
                                                group_paths,
                                                group_dicts,
                                                )
                                            ))(dataflow_config)

    return updated_dataflow_config


if __name__ == '__main__':
    register_yaml_schemas(path='configs/schemas/base/')
    print()



"""
Setup script for the Quinine library.
"""
import os
from setuptools import setup, find_packages


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


def req_file(filename):
    with open(filename) as f:
        content = f.readlines()
    return [x.strip() for x in content]


install_requires = req_file("requirements.txt")

setup(
    name="quinine",
    version="0.2.0",
    author="Karan Goel",
    author_email="kgoel93@gmail.com",
    license="MIT",
    description="quinine is a library for configuring machine learning projects.",
    keywords="configuration yaml machine learning ml ai nlp cv vision deep learning",
    # url="http://packages.python.org/an_example_pypi_project",
    packages=['quinine', 'quinine.common'],
    # long_description=read('README.md'),
    install_requires=install_requires,
)
